// rewritten by Zheng Gengbin
//
// add translation of realign and redistribute.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
extern "C" void exit(...);
#endif

#include <stdio.h>
#include "compilerApi.h"
#include "strategy.h"
#include "anls_att.h"

extern Shadow_def * get_defined_shadow(SgSymbol *);

extern Flag clock_flag;

//extern "C" int strncasecmp(...);

void add_include(SgStatement * s)
{

  SgStatement * firstdecl=s->childList1(0);
  
  // add stack size def
  {
#ifndef USE_F90
  SgValueExp * filename = new SgValueExp("pcrc.inc");
#else
  SgValueExp * filename = new SgValueExp("pcrc90.inc");
#endif
  SgStatement * include = new SgStatement(INCLUDE_STAT, NULL, NULL, filename, NULL, NULL);

  firstdecl->insertStmtBefore(*include, *s);
  }

  // add new def
  {
#ifdef MEM_ALLOC
  SgValueExp * filename = new SgValueExp("mafdecls.inc");
#else
  SgValueExp * filename = new SgValueExp("pcrc_stack.inc");
#endif
  SgStatement * include = new SgStatement(INCLUDE_STAT, NULL, NULL, filename, NULL, NULL);

  firstdecl->insertStmtBefore(*include, *s);
  }
}


void add_init_finalize(SgStatement * s, SgStatement * lastdec, SgStatement * lastexec)
{
  int idsize;
  SgStatement *cp = lastdec?lastdec->controlParent():s;
  SgStatement *where = lastdec?lastdec:s;

  {
#ifdef MEM_ALLOC
    {
    SgSymbol *fname = s->GetSymbolInScope(cp, mixstring, "mem", "init");
    SgCallStmt * afuncall = new SgCallStmt(*fname);
    if (stack_size > 0) {
    	stack_size = stack_size<MIN_STACK_SIZE?MIN_STACK_SIZE:stack_size;
    	stack_size = stack_size>MAX_STACK_SIZE?MAX_STACK_SIZE:stack_size;
    }
    else {
	stack_size = -stack_size;
    }
    SgValueExp *stacksize = new SgValueExp(stack_size);
    afuncall->addArg(*stacksize);

    SgValueExp *heapsize = new SgValueExp(heap_size);
    afuncall->addArg(*heapsize);
    where->insertStmtAfter(*afuncall, * cp);

    }
#endif

    SgStatement *afuncall = make_init(cp, storageMode, runtimeDebug);
    where->insertStmtAfter(*afuncall, * cp);

    if ( clock_flag == on )
    {
    SgStatement *afuncall = make_clock(cp, 0);
    where->insertStmtAfter(*afuncall, * cp);
    }
  };

  {
    if ( clock_flag == on )
    {
    SgStatement *afuncall = make_clock(cp, 1);
    lastexec->insertStmtBefore(*afuncall, * cp);
    }

    SgSymbol * fname = s->GetSymbolInScope(cp, mixstring, "finalize", NULL);
    SgStatement * afuncall = new SgCallStmt(*fname);
    lastexec->insertStmtBefore(*afuncall, * cp);

  };

}

void add_stack_operation(SgStatement * s, SgStatement * lastdec, SgStatement * lastexec)
{
  SgStatement *cp = lastdec?lastdec->controlParent():s;
  SgStatement *where = lastdec?lastdec:s;

  SgProcHedrStmt * prochedr = (SgProcHedrStmt *) s;

  for (int i = 0; i<prochedr->numberOfParameters(); i++) { 
    SgSymbol * parametersymb = prochedr->parameter(i);

//   I think no matter Inherit_Def , we always add reset and pop
//    if (parametersymb->numberOfAttributes(Inherit_Def)!=0) {

      SgType * parametertype = parametersymb->type();
      if (parametertype->variant()==T_ARRAY) {

	SgArrayType * ptype = (SgArrayType *) parametertype;

	for (int j = 0; j<ptype->dimension(); j++) {
	  // reset_push_bounds
	  {
	    SgSymbol * fname = s->GetSymbolInScope(cp, mixstring, "reset_push_bounds", NULL);
	    SgCallStmt * afuncall = new SgCallStmt(*fname);

	    SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", parametersymb->identifier());
	    SgExpression * dadexp = new SgVarRefExp ( dad);
	    afuncall->addArg(* dadexp);

	    SgValueExp * dim = new SgValueExp(j+1);
	    afuncall->addArg(* dim);

	    SgValueExp * start;

	    SgTripletExp * size = (SgTripletExp * ) ptype->sizeInDim(j);

	    if (size->numberOfItems()!=0)
              // modified by zheng.gb   1996/11/22
              //start = new SgValueExp (size->item(0)->valueInteger());
	      start = new SgValueExp(size->item(0)?size->item(0)->valueInteger():1);
	    else
	      start = new SgValueExp (1);

	    
	    afuncall->addArg(* start);
	    
	    //SgValueExp * end = new SgValueExp (size->item(1)->valueInteger());
	    //afuncall->addArg(* end);

	    SgSymbol * actlb = s->getSymbolInCurrentScope(mixstring, "alb", parametersymb->identifier(), j);
	    SgExpression * actlbexp = new SgVarRefExp (actlb);

	    afuncall->addArg( *actlbexp);

	    SgSymbol * actub = s->getSymbolInCurrentScope(mixstring, "aub", parametersymb->identifier(), j);
	    SgExpression * actubexp = new SgVarRefExp (actub);

	    afuncall->addArg( *actubexp);

	    SgSymbol * actstr = s->getSymbolInCurrentScope(mixstring, "ast", parametersymb->identifier(), j);
	    SgExpression * actstrexp = new SgVarRefExp (actstr);

	    afuncall->addArg( *actstrexp);

	    where->insertStmtAfter(*afuncall, *cp);
	  };

	  //reset_pop_bounds
	  {
	    SgSymbol * fname = s->GetSymbolInScope(cp, mixstring, "reset_pop_bounds", NULL);
	    SgCallStmt * afuncall = new SgCallStmt(*fname);
	    
	    SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", parametersymb->identifier());
	    SgExpression * dadexp = new SgVarRefExp (dad);
	    afuncall->addArg(* dadexp);

	    SgValueExp * dim = new SgValueExp(j+1);
	    afuncall->addArg(* dim);


	    SgSymbol * actlb = s->getSymbolInCurrentScope(mixstring, "alb", parametersymb->identifier(), j);
	    SgExpression * actlbexp = new SgVarRefExp (actlb);

	    afuncall->addArg( *actlbexp);

	    SgSymbol * actub = s->getSymbolInCurrentScope(mixstring, "aub", parametersymb->identifier(), j);
	    SgExpression * actubexp = new SgVarRefExp (actub);

	    afuncall->addArg( *actubexp);

	    SgSymbol * actstr = s->getSymbolInCurrentScope(mixstring, "ast", parametersymb->identifier(), j);
	    SgExpression * actstrexp = new SgVarRefExp (actstr);

	    afuncall->addArg( *actstrexp);
	
	    lastexec->insertStmtBefore(*afuncall, * lastexec->controlParent());
	  };
	}       // end of for
      }
    }           // end of if Inherit_Def
  }             // end of for
//}             // end of if Inherit_Def


void define_align_collapse(Alignment_def *adef , int dim ,SgStatement *s,SgStatement *lastdec)
{
      SgStatement *cp = s->getScopeForDeclare();
      SgSymbol *arraySymb = adef->getfromsymbol();
      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraySymb->identifier());

      SgExpression * start = get_lbound(adef->getfromsymbol(),dim);
      SgExpression * end =  get_ubound(adef->getfromsymbol(),dim);

      SgStatement *afuncall =  make_set_array_collapse(s, dad, dim+1, start, end);

      lastdec->insertStmtAfter(*afuncall, *cp);

	// in case start and end have intrinsic functions.
      analyse_expression(afuncall, start);
      analyse_expression(afuncall, end);
}

void define_align_array(SgStatement * s, SgStatement * lastdec, SgStatement * lastexec)
{
  SgStatement *cp = s->getScopeForDeclare();

  for (int i=0; i<s->numberOfAttributes(Align_Def); i++) {
    Alignment_def * adef = (Alignment_def *)s->attributeValue(i, Align_Def);

    SgSymbol *arraySymb = adef->getfromsymbol();
    int storage = getArrayStorage(arraySymb);

// for out of core array, add the initialization statement
// pcrc_oc_init(dad,incoresize) 
    if ( is_out_of_core(adef->getfromsymbol()) )
    {
      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraySymb->identifier());
      SgStatement * ts = make_oc_init(s, dad, IN_CORE_SIZE);
      lastdec->insertStmtAfter(*ts, *cp);
    } // end of Outofcore_Def

    //for each dim, add set_array_rang_align;
    // duplicated(*) in align_with not considered !!!

    int j;
    for (int dim=adef->getDim()-1; dim>=0; dim--)  {
      if ( storage == COLUMN_MAJOR )
	j = dim;
      else
	j = adef->getDim() - dim - 1;

      int to_dim = adef->getAlignPair(j)+1;

      // if is collaspe
      if (to_dim == 0) {
	define_align_collapse(adef,j,s,lastdec);
	continue;
      }

        // void pcrc_set_array_align(int* dad, int* dim, int* g_lb, int* g_ub, int* a, int* b, int* rng_parent)
      SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "set_array_align", NULL);
      SgCallStmt * afuncall = new SgCallStmt(*fname);

      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraySymb->identifier());
      SgExpression * expr = new SgVarRefExp(dad);

      afuncall->addArg(*expr);

      SgValueExp * dim = new SgValueExp(j+1);
      afuncall->addArg(*dim);

      SgExpression * start=get_lbound(arraySymb,j);
      afuncall->addArg(*start);
      
      SgExpression * end =  get_ubound(arraySymb,j);
      afuncall->addArg(*end);

      SgExpression * stride = adef->getStride(j);
      afuncall->addArg(*stride);

      SgExpression * offset = adef->getOffset(j);
      analyse_expression(lastdec, offset);
      afuncall->addArg(*offset);      
     
      if (adef->getfromsymbol()->numberOfAttributes(Shad_Def)!=0) {
	Shadow_def * sdef=NULL;
	
	for (int k=0; k<adef->getfromsymbol()->numberOfAttributes(Shad_Def); k++) {
	  Shadow_def * gsdef=(Shadow_def *)arraySymb->attributeValue(k, Shad_Def);
	  if (gsdef->getTag()==GNEEDED) {
	    sdef=gsdef;
	    break;
	  }
	}
	ASSERT(sdef);
	
	SgValueExp * gtl = new SgValueExp(sdef->getLower(j));
	SgValueExp * gtu = new SgValueExp(sdef->getUpper(j));

	afuncall->addArg(*gtl);      
	afuncall->addArg(*gtu);      		

      }
      else {
	SgValueExp * zero = new SgValueExp(0);
	afuncall->addArg(*zero);      
	afuncall->addArg(*zero);      	
      }
      
      SgSymbol * tmpsymb = adef->gettosymbol();

      if ( tmpsymb->numberOfAttributes(Temp_Def) == 0)   {
	// then tmpsymb must be a dummy with inherit             zheng.gb
	// simply call the function  pcrc_range() to got the range
//printf("Call pcrc_rang() to symbol %s with inherit attribute .\n",adef->getfromsymbol()->identifier());
        SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", tmpsymb->identifier());
      	SgValueExp * dimention = new SgValueExp(to_dim);
	SgExpression *rngfunc = make_range(s,dad,dimention);
      	afuncall->addArg(*rngfunc);
      }
      else {
      	Template_def * tdef = (Template_def *)tmpsymb->attributeValue(0, Temp_Def);
      	SgSymbol * rngsymb = tdef->getrng();
	  // realign will be done before add_template_declaration
        if (rngsymb == NULL) {
	  //ASSERT(is_dynamic(tmpsymb));
    	  rngsymb = s->getSymbolInCurrentScope(mixstring, "rng", tdef->getsymbol()->identifier());
        }
      
      	SgValueExp * dimention = new SgValueExp(to_dim);
      	SgExpression * range = new SgExpression(RANGE_OP, NULL, dimention, NULL);
      	SgExpression * ddot = new SgExpression(DDOT, range, NULL, NULL);
      	SgExpression * index = new SgExpression(EXPR_LIST, ddot, NULL, NULL);
      	SgExpression * rngarray = new SgExpression(ARRAY_REF, index, NULL, rngsymb);
      	afuncall->addArg(*rngarray);
      }

      lastdec->insertStmtAfter(*afuncall, *cp);

	// in case start and end have intrinsic functions.
      analyse_expression(afuncall, start);
      analyse_expression(afuncall, end);
    }

    // new_array_data
    {
      SgType * stype= arraySymb->type();

      SgExpression *address;
      if (arraySymb->numberOfAttributes(UNBOUND_ARRAY)  == 0)
	 address = new SgVarRefExp(arraySymb);
      else {
	 SgSymbol *stack = s->getSymbolInCurrentScope(stackNameOfType(stype),"mb", NULL);
	 address = new SgVarRefExp(stack);
      }

      // in the following we want the grp name of the prcossor;

      SgSymbol * templatename= adef->gettosymbol();
      SgExpression *grpExp = NULL;
      if ( templatename->numberOfAttributes(Dist_Def) == 0)   {
	  // I assume then templatename must be a dummy with inherit   zheng.gb
	  // simply call the function  pcrc_group() to got the group
        SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", templatename->identifier());
	grpExp = make_group(s, dad);
      }
      else {
      	Distribute_def *ddef = (Distribute_def *)templatename->attributeValue(0, Dist_Def);

/*
      	SgSymbol * procname = ddef->gettosymbol();
      	Processor_def * pdef = (Processor_def *)procname->attributeValue(0, Proc_Def);
      	SgSymbol * procgrd = pdef->getgrp();
*/
      	SgSymbol * procgrd = ddef->getgrp();
      	grpExp = new SgVarRefExp(procgrd);
      }

      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraySymb->identifier());

      SgStatement *assign = make_new_array_data1(s, dad, address, stype, storage,
						adef->getDim(),grpExp);

      lastdec->insertStmtAfter(*assign, *cp);
    }

    // below is the couter part of the above set new array;
    if (lastexec)
    {
      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraySymb->identifier());
      SgStatement *afuncall =  make_delete_array(s , dad);
      lastexec->insertStmtBefore(*afuncall, *cp);
    }
  };
}

// for array which is of unknown array bound, allocate memory
void allocate_unbound_array(SgStatement * s, SgStatement * lastdec, SgStatement * lastexec)
{
  SgStatement *cp = s->getScopeForDeclare();

  for (int i=0; i<s->numberOfAttributes(Align_Def); i++) 
  {
    Alignment_def * adef = (Alignment_def *)s->attributeValue(i, Align_Def);

    SgSymbol *arraySymb = adef->getfromsymbol();

    if (is_dummy(s, arraySymb)) continue;

    // we only deal with UNBOUND_ARRAY / DYNAMIC
    int skip = 1;
    if (arraySymb->numberOfAttributes(UNBOUND_ARRAY) > 0) skip = 0;
    if (is_dynamic(arraySymb)) skip = 0;
    if (skip == 1) continue;

    int need_heap = 0;
    if (is_dynamic(arraySymb))  need_heap = 1;

    SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraySymb->identifier());

    SgSymbol * mHandle;
#ifdef MEM_ALLOC
     mHandle = s->getSymbolInCurrentScope(mixstring, "mh", NULL, 0);
#endif
     SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "tbs", NULL, 0);

    // reverse seqeunce
    // reset base;
    {
    SgSymbol * basesym = s->getSymbolInCurrentScope(mixstring, "tbs", NULL,0);
    SgStatement *funcall=make_reset_array_base(s,dad,arraySymb->type(),basesym, mHandle);
    lastdec->insertStmtAfter(*funcall, *cp);
    }

     // allocate memory;
     // allocate memory according to RHS symbol's type, even if create tdd as LHS (flag==0)
    {
    SgStatement *alloc;
    if (need_heap = 0)
    {
#ifdef MEM_ALLOC
	alloc=make_stack_alloc(s,arraySymb->type(),base,dad,mHandle);
#else
	alloc=make_alloc(s,rhsarraysymb->type(),base,dad);
#endif
    }
    else  {
	alloc=make_heap_alloc(s,arraySymb->type(),base,dad,mHandle);
    }
    lastdec->insertStmtAfter(*alloc, *cp);
    }

    // call real free;
    {
    SgStatement* free_stmt = make_array_free(s, dad);
    lastexec->insertStmtBefore(*free_stmt, *lastexec->controlParent());
    }
  }
}

void define_noalign_data(SgStatement * s, SgStatement * lastdec, SgStatement * lastexec)
{
  SgStatement *cp = s->getScopeForDeclare();
  SgSymbol * first = ((SgProgHedrStmt *)s)->firstSymbolInScope();

  for (; first; first=first->nextInScope()) {
    if (first->numberOfAttributes(NeedDad_Def)!=0) {
      SgSymbol * ndsymb= first;
      SgType * stype= ndsymb->type();

      int storage = getArrayStorage(ndsymb);

      int dim=0;
      if (isSgArrayType(stype))
    	   dim = ((SgArrayType *)stype)->dimension();
	
      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", ndsymb->identifier());

	// collapse each dim if is an array;
      {
        int j;
        for (int d=dim-1; d>=0; d--)  {
          if ( storage == COLUMN_MAJOR )
		j = d;
          else
		j = dim - d - 1;

          SgExpression * start = get_lbound(ndsymb,d);

          SgExpression * end =  get_ubound(ndsymb,d);

          SgStatement *afuncall =  make_set_array_collapse(s, dad, d+1, start, end);

          lastdec->insertStmtAfter(*afuncall, *cp);

          analyse_expression(afuncall, start);
          analyse_expression(afuncall, end);
        }
      }

      {
	SgExpression *address = new SgVarRefExp(ndsymb);

        SgStatement *assign = make_new_array_data1(s,dad,address,stype,
                                                  storage, dim, NULL);
	
	lastdec->insertStmtAfter(*assign, *s);
      }	
        // below is the couter part of the above set new array;
      if (lastexec)
      {
        SgStatement *afuncall =  make_delete_array(s , dad);
        lastexec->insertStmtBefore(*afuncall, *cp);
      }
    } // if
  }
}

void make_sub_grid(SgStatement *s, SgSymbol *subgrp, Processor_def *pdef, SgArrayRefExp *procs, int way)
{
      int rank = pdef->getDim();

      SgSymbol *grp = s->getSymbolInCurrentScope(mixstring, "grp", pdef->getsymbol()->identifier());
      SgSymbol *symb1=s->getSymbolInCurrentScope(mixstring, "a0", NULL);
      SgSymbol *symb2=s->getSymbolInCurrentScope(mixstring, "b0", NULL);
      SgSymbol *symb3=s->getSymbolInCurrentScope(mixstring, "c0", NULL);

      SgStatement *subgrid_stmt = make_sub_group_grid(s, subgrp, rank, 
						symb1, symb2, symb3, grp);

      if (way == AFTER) {
      	s->insertStmtAfter(*subgrid_stmt, *s->controlParent());
      }

      for (int i=0; i<rank; i++)
      {
	SgExpression *sub = new SgValueExp(i+1);
	SgArrayRefExp *arrayExp1 = new SgArrayRefExp(*symb1);
        arrayExp1->addSubscript(*sub);
	SgArrayRefExp *arrayExp2 = new SgArrayRefExp(*symb2);
        arrayExp2->addSubscript(*sub);
	SgArrayRefExp *arrayExp3 = new SgArrayRefExp(*symb3);
        arrayExp3->addSubscript(*sub);

	SgExpression *rhsExp1 = get_lbound_from_arrayexp(procs, i);
        ASSERT(rhsExp1);
	SgExpression *rhsExp2 = get_ubound_from_arrayexp(procs, i);
        ASSERT(rhsExp2);
	SgExpression *rhsExp3 = get_stride_from_arrayexp(procs, i);
        ASSERT(rhsExp3);

	SgAssignStmt *assign1 = new SgAssignStmt(*arrayExp1,*rhsExp1);
        SgAssignStmt *assign2 = new SgAssignStmt(*arrayExp2,*rhsExp2);
        SgAssignStmt *assign3 = new SgAssignStmt(*arrayExp3,*rhsExp3);
        if (way == BEFORE) {
          s->insertStmtBefore(*assign1, *s->controlParent());
          s->insertStmtBefore(*assign2, *s->controlParent());
          s->insertStmtBefore(*assign3, *s->controlParent());
        }
	else {
          s->insertStmtAfter(*assign3, *s->controlParent());
          s->insertStmtAfter(*assign2, *s->controlParent());
          s->insertStmtAfter(*assign1, *s->controlParent());
	}
      }
      if (way == BEFORE) {
      	s->insertStmtBefore(*subgrid_stmt, *s->controlParent());
      }
}


// s is a proc_hedr
void distribute_array(SgStatement *s,SgSymbol *fromdad,SgExpression *arrayexp,Distribute_def *ddef,Processor_def *pdef,SgStatement *where,int way)
{
      SgStatement * afuncall;
        // if from symbol cannot get size , it must have dummy in it , use the size of arrayexp instead
      int sizable=1;
      if (arrayexp) {
	SgArrayType *arraytype = isSgArrayType(ddef->getfromsymbol()->type());
	if (arraytype && arraytype->getSize() == -1) sizable = 0;
      }

        // get the dad name of distribute
      SgSymbol * dad ;
      if ( fromdad ) 
	dad = fromdad;
      else
      	dad = s->getSymbolInCurrentScope(mixstring, "dad", ddef->getfromsymbol()->identifier());

      SgSymbol *arraysymb = ddef->getfromsymbol();
      ASSERT(arraysymb);

	// add by slx 98/3/10
        // if outofcore, pcrc_oc_init(pcrc_tdd,IN_CORE_SIZE,MAJOR)
      {
   	 if(arraysymb->numberOfAttributes(Outofcore_Def) > 0 &&
	    way == AFTER)  
     	 {
      	  SgStatement *funcall = make_oc_init(where, dad, IN_CORE_SIZE);
      	  where->insertStmtAfter(*funcall, *where->controlParent());
     	 }
      }  

      int dim,j;

	// have SHADOW
      Shadow_def *sdef = NULL;
      if (arraysymb->numberOfAttributes(Shad_Def) > 0)
      {
	  // retreive defined shadow
	sdef = get_defined_shadow(arraysymb);
      }

      int storage = getArrayStorage(arraysymb);

      int arraydim = 0;
      SgArrayRefExp *array = isSgArrayRefExp(arrayexp);
      if(array)
      {
      for (j = 0; j<array->numberOfSubscripts(); j++)
      	if (!array->isScalarOnDim(j)) arraydim ++;
      ASSERT(arraydim == ddef->getDim());
      }

      if ( sdef != NULL )
      {
	  // create temporary Range class
      	SgSymbol *rngsymb = s->getSymbolInCurrentScope(mixstring, "tmp_rng", ddef->getfromsymbol()->identifier());

        for (dim = 0; dim < ddef->getDim() ; dim++ )
	{
	    SgStatement *stmt[3];

	    // ddim is the dimension of array distributing(dummy)
	    // j is the dimension of arraysymb(fromsymb)
	    int ddim, d;
	    if ((storage == COLUMN_MAJOR && way == BEFORE) ||
		(storage == ROW_MAJOR && way == AFTER) )
		ddim = dim;
	    else
		ddim = ddef->getDim() - dim -1;
	    j = ddim;

		// search the array dimension(j) if is an array
	    if (array)
	    {
	      for (d = ddim, j = 0 ; j< array->numberOfSubscripts(); j++)
	      {
		if ( array->isScalarOnDim(j) ) continue;
		if (d == 0) break;
		d --;
	      }
	    }

      	    SgArrayRefExp *rngarray = new SgArrayRefExp(*rngsymb);
            SgValueExp * dimention = new SgValueExp(j+1);
            SgTripletExp *ddot = new SgTripletExp(NULL,dimention,NULL);
            rngarray->addSubscript(*ddot);

	      // get the lbound and ubound
	    SgExpression *start , *end;
	    if (sizable == 1)
	    {
      	    	start = get_lbound(ddef->getfromsymbol(),ddim);
      		end =  get_ubound(ddef->getfromsymbol(),ddim);
	    }
	    else 
	    {
		start = get_lbound_from_arrayexp(arrayexp , j);
		end = get_ubound_from_arrayexp(arrayexp , j);
		SgExpression *str = get_stride_from_arrayexp(arrayexp, j);
		SgExpression *v1  = new SgValueExp(1);
		end = &( (*end - *start->copyPtr() + *v1 )/ *str);
		analyse_expression(s, start);
		analyse_expression(s, end);
	    }

	      // new range
	    int distcode = ddef->distCode(j);
	    if (distcode == NOP ) {
	        SgExpression * collapse = make_new_range_collapse(s,start,end);
		stmt[0] = new SgAssignStmt(*rngarray, *collapse);
	    }
	    else {
		switch (distcode) {
		case BLOCK:  distcode = 1 ; break;
		case CYCLIC: distcode = 2 ; break;
		}

	    	SgSymbol * procgrd = pdef->getgrp();
	    
	    	// we need the processor dim here;
	    	int processorDimCounter=0;
	    	for (int k=0; k<=j; k++) {
	      	if (ddef->distCode(k)!=NOP)
			processorDimCounter++;
	    	}
		SgExpression *distribute = make_new_range_distribute(s,start,end ,
					j+1 ,procgrd ,processorDimCounter );
		stmt[0] = new SgAssignStmt(*rngarray, *distribute);
	    }

		// set_array_align
	    {
	    SgExpression *a = new SgValueExp(1);
	    SgExpression *b = new SgValueExp(0);
	    SgExpression *gl = new SgValueExp(sdef->getLower(j));
	    SgExpression *gu = new SgValueExp(sdef->getUpper(j));

	    stmt[1] = make_set_array_align(s,dad,ddim+1,start,end,a,b,gl,gu,rngarray);
            }

 	      // delete range 
	    {
	    stmt[2] = make_delete_range(s, rngarray);
	    }

	    for (int i=0; i<3; i++)
	    if (way == AFTER)
      		where->insertStmtAfter(*stmt[2-i], *where->controlParent());
	    else
      		where->insertStmtBefore(*stmt[i], *where->controlParent());

	}  // end of for
      }
      else 
      {
	  // distribute array, set dim information to dad
        for (dim = 0; dim < ddef->getDim() ; dim++)
	{
	    // ddim is the dimension of array distributing
	    // j is the dimension of arraysymb(fromsymb)
	    int ddim, d;
	    if ((storage == COLUMN_MAJOR && way == BEFORE) ||
		(storage == ROW_MAJOR && way == AFTER) )
	    	ddim = dim;
	    else
		ddim = ddef->getDim() - dim -1;
	    j = ddim;

	      // search the array dimension(j) if is an array, reset (j)
	    if (array)
	    {
	      for (d= ddim, j = 0 ; j< array->numberOfSubscripts(); j++)
	      {
		if ( array->isScalarOnDim(j) ) continue;
		if (d == 0) break;
		d --;
	      }
	    }

	      // get the lbound and ubound
	    SgExpression *start , *end;
	    if (sizable == 1)
	    {
                start = get_lbound(ddef->getfromsymbol(),ddim);
      		end =  get_ubound(ddef->getfromsymbol(),ddim);
	    }
	    else 
	    {
      	        start = get_lbound_from_arrayexp(arrayexp,j); ASSERT(start);
		end  = get_ubound_from_arrayexp(arrayexp, j);
		SgExpression *str = get_stride_from_arrayexp(arrayexp, j);
		SgExpression *v1  = new SgValueExp(1);
		end = &( (*end - *start->copyPtr() + *v1 )/ *str);
	    }

	    int distcode = ddef->distCode(j);
	    if (distcode == NOP ) 
	    {
		// collapse
		afuncall = make_set_array_collapse(where, dad, ddim+1, start, end);
	    }
	    else {
		switch (distcode) {
		case BLOCK:  distcode = 1 ; break;
		case CYCLIC: distcode = 2 ; break;
		}

	    	SgSymbol * procgrd = pdef->getgrp();
	    
	    	// we need the processor dim here;
	    	int processorDimCounter=0;
	    	for (int k=0; k<=j; k++) {
	      	if (ddef->distCode(k)!=NOP)
			processorDimCounter++;
	    	}

		afuncall = make_set_array_distribute(s, dad, ddim+1, start, end, 
                                       distcode, procgrd, processorDimCounter);

	     }
 	     if (way == AFTER)
      		where->insertStmtAfter(*afuncall, *where->controlParent());
	     else
      		where->insertStmtBefore(*afuncall, *where->controlParent());
	     analyse_expression(afuncall, start);
	     analyse_expression(afuncall, end);
        }	// end of for (j)
      }

	// add by slx 98/3/10
	// if outofcore, pcrc_oc_init(pcrc_tdd,IN_CORE_SIZE,MAJOR)
      {
   	 if(arraysymb->numberOfAttributes(Outofcore_Def) > 0 &&
	    way == BEFORE )  
     	 {
      	  SgStatement *funcall = make_oc_init(where, dad, IN_CORE_SIZE);
      	  where->insertStmtBefore(*funcall, *where->controlParent());
     	 }
      }  
}

// statement s is a proc_hedr
void define_distribute_array(SgStatement *s , SgStatement *lastdec , SgStatement *lastexec)
{
  for (int i=0; i<s->numberOfAttributes(DDist_Def); i++) 
  {
	Distribute_def * ddef = (Distribute_def *) s->attributeValue(i,DDist_Def);
	SgSymbol * procname = ddef->gettosymbol();
    	Processor_def * pdef = (Processor_def *)procname->attributeValue(0, Proc_Def);

	distribute_array(s,NULL,NULL,ddef,pdef,lastdec,AFTER);
  }
}

void distribute_template(SgStatement *s, SgSymbol * rngsymb, Template_def *tdef, Distribute_def *ddef, SgStatement *where)
{
      SgStatement *cp = s->getScopeForDeclare();
      SgSymbol * procname = ddef->gettosymbol();
      Processor_def * pdef = (Processor_def *)procname->attributeValue(0, Proc_Def);

      SgArrayRefExp *subProc = ddef->getProcs();
        // in the following we want the grp name of the prcossor;
      SgSymbol * procgrd = ddef->getgrp();	// pdef->getgrp()
      ASSERT(procgrd);
/*
      SgSymbol * procgrd = pdef->getgrp();
*/

      for (int j=tdef->getDim()-1; j>=0; j--) 
      {
	  // LHS of assign
      	SgArrayRefExp *rngarray = new SgArrayRefExp(*rngsymb);
        SgValueExp * dimention = new SgValueExp(j+1);
        SgTripletExp *ddot = new SgTripletExp(NULL,dimention,NULL);
        rngarray->addSubscript(*ddot);

	switch (ddef->distCode(j)) {
	case NOP:
	  {
	    SgExpression * start= tdef->getStart(j);
	    SgExpression * end = tdef->getEnd(j);
	    SgExpression *funcexpr = make_new_range_collapse(s,start,end);

	    SgAssignStmt * assign = new SgAssignStmt(*rngarray, *funcexpr);
	    
	    where->insertStmtAfter(*assign, *cp);
	    break;
	  }
	case CYCLIC:
	case BLOCK: 
	  {
	    SgExpression * start= tdef->getStart(j);
	    SgExpression * end = tdef->getEnd(j);

	    // we need the processor dim here;
	    int processorDimCounter=0;
	    for (int k=0; k<=j; k++) {
	      if (ddef->distCode(k)!=NOP)
		processorDimCounter++;
	    }

	    int distcode = ddef->distCode(j) == BLOCK ? 1:2;
	    SgExpression *funcexpr = make_new_range_distribute(s,start,end,
				          distcode,procgrd,processorDimCounter);
	    
	    SgAssignStmt * assign = new SgAssignStmt(*rngarray, *funcexpr);
	    
	    where->insertStmtAfter(*assign, *cp);
	    break;
	  }
	default:
	    ASSERT(0);
	    break;
	}
      }

      if (subProc && isSgArrayRefExp(subProc)) {
      	  SgSymbol * procname = ddef->gettosymbol();
      	  Processor_def *pdef = (Processor_def *)procname->attributeValue(0, Proc_Def);

    	  SgArrayRefExp *procs = ddef->getProcs();
          make_sub_grid(where, procgrd, pdef, procs, AFTER);
      }
}

void add_template(SgStatement * s, SgStatement * lastdec, SgStatement * lastexec)
{
  
  for (int i=0; i<s->numberOfAttributes(Temp_Def); i++) {
    
    Template_def * tdef = (Template_def *)s->attributeValue(i, Temp_Def);
    SgSymbol * templatename= tdef->getsymbol();    
    SgSymbol * rngsymb = tdef->getrng();
    Distribute_def * ddef = (Distribute_def *)templatename->attributeValue(0, Dist_Def);

      // create template as ddef describe
    distribute_template(s, rngsymb, tdef, ddef, lastdec);

    if (lastexec)
    {
      for (int j=tdef->getDim()-1; j>=0; j--) 
      {
      
        SgArrayRefExp *rngarray = new SgArrayRefExp(*rngsymb);
	SgValueExp *dimention = new SgValueExp(j+1);
        SgTripletExp *ddot = new SgTripletExp(NULL,dimention,NULL);
        rngarray->addSubscript(*ddot);

	SgStatement *afuncall = make_delete_range(s,rngarray);

	lastexec->insertStmtBefore(*afuncall, *s);
      }
    }

  }
}

// add processor definition in the scope of header s , after or before where1.
// add a delete_group after or before where2 
void add_a_processor(SgStatement *s , Processor_def *pdef , SgStatement *where1 , int way1, SgStatement *where2,int way2)
{
    {
      SgExpression *funcexpr=make_new_group_grid(s,pdef->getDim(),pdef->getshp());

      SgSymbol * procgrp = pdef->getgrp();
      SgExpression * procgrpexpr = new SgVarRefExp(procgrp);
      
      SgAssignStmt * assign = new SgAssignStmt(*procgrpexpr, *funcexpr);

      where1->insertStmtAfter(*assign, *where1->controlParent()); //  s =>  zheng.gb
    };
    
    for (int j =0; j< pdef->getDim(); j++) {
      SgExpression *start = pdef->getStart(j)->copyPtr();
      analyse_expression(s, start);  // check intrinsic functions
      SgExpression *end = pdef->getEnd(j)->copyPtr();
      analyse_expression(s, end);  // check intrinsic functions
      SgExpression * dimsize= &(*end - *start + *new SgValueExp(1) );
      dimsize = calculateExp(dimsize);
      
      SgArrayRefExp *shparray = new SgArrayRefExp(*pdef->getshp());
      SgValueExp * dimention = new SgValueExp(j+1);
      SgTripletExp *ddot = new SgTripletExp(NULL,dimention,NULL);
      shparray->addSubscript(*ddot);

      SgAssignStmt * assign = new SgAssignStmt(*shparray, *dimsize);
      
      where1->insertStmtAfter(*assign, *where1->controlParent());  // s=>  zheng.gb
    };
    {
      SgStatement *afuncall = make_delete_group(s , pdef->getgrp());
      where2->insertStmtBefore(*afuncall, *where2->controlParent());  // s=> zheng.gb
    };
}

void add_processor(SgStatement * s, SgStatement * lastdec, SgStatement * lastexec)
{
  for (int i=0; i<s->numberOfAttributes(Proc_Def); i++) {
    Processor_def * pdef = (Processor_def *)s->attributeValue(i, Proc_Def);

    {
      SgExpression *funcexpr=make_new_group_grid(s,pdef->getDim(),pdef->getshp());

      SgSymbol * procgrp = pdef->getgrp();
      SgExpression * procgrpexpr = new SgVarRefExp(procgrp);
      
      SgAssignStmt * assign = new SgAssignStmt(*procgrpexpr, *funcexpr);

      lastdec->insertStmtAfter(*assign, *lastdec->controlParent()); //  s =>  zheng.gb
    };
    
    for (int j =0; j< pdef->getDim(); j++) {
      SgExpression * dimsize= &(*pdef->getEnd(j) - *pdef->getStart(j) + *new SgValueExp(1));
      analyse_expression(s, dimsize);  // check intrinsic functions
      dimsize = calculateExp(dimsize);

      SgArrayRefExp *shparray = new SgArrayRefExp(*pdef->getshp());
      SgValueExp * dimention = new SgValueExp(j+1);
      SgTripletExp *ddot = new SgTripletExp(NULL,dimention,NULL);
      shparray->addSubscript(*ddot);

      SgAssignStmt * assign = new SgAssignStmt(*shparray, *dimsize);
      
      lastdec->insertStmtAfter(*assign, *lastdec->controlParent());  // s=>  zheng.gb
    };
    {
      SgStatement *afuncall = make_delete_group(s , pdef->getgrp());
      lastexec->insertStmtBefore(*afuncall, *lastexec->controlParent());  // s=> zheng.gb
    };

  }
}

void define_return_label(SgStatement * s)
{
  if (s->numberOfAttributes(RETURN_LABEL)!=0) 
  {
    SgStatement *lastexec=s->childList1(s->numberOfChildrenList1()-1);
    SgStatement * continueStmt = new SgStatement(CONTINUE_NODE);
    SgLabel *label = new SgLabel(*continueStmt);
    continueStmt->setLabel(*label);

    lastexec->insertStmtBefore(*continueStmt, * lastexec->controlParent());
    s->deleteAllAttribute(RETURN_LABEL);
    s->addAttribute(RETURN_LABEL, label, sizeof(label));
  }
}

void define_stop_label(SgStatement * s)
{
  define_return_label(s);
}

/***************************************************************************
                       REALIGN and REDISTRIBUTE

	void realign_array(SgStatement *s)
	void redistribute_array(SgStatement *s)
***************************************************************************/

//  realign statement
// TODO memory is not freed
void realign_array(SgStatement *s)
{
   SgSymbol *symb = s->expr(0)->lhs()->symbol();
   SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", symb->identifier());
   SgSymbol * sdad = s->getSymbolInCurrentScope(mixstring, "sdd", symb->identifier());

   int allocate = 0;
   if (symb->numberOfAttributes(Allocate_Def)>0) allocate=1;

   // create statements inversely

   if (allocate == 0)
   {
     // delete old DAD
   SgStatement *deleteStmt = make_delete_array(s, sdad);
   s->insertStmtAfter(*deleteStmt, *s->controlParent());

     // delete the allocated memory of old dad
   SgStatement *freeStmt = make_array_free(s, sdad);
   s->insertStmtAfter(*freeStmt, *s->controlParent());

     // add remap
   SgStatement *remap = make_remap(s, dad, sdad);
   s->insertStmtAfter(*remap, *s->controlParent());
   }

   SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "tbs", symb->identifier(), 0);
   SgSymbol * mHandle = s->getSymbolInCurrentScope(mixstring, "mh", symb->identifier(), 0);

    // reset_array_base
   SgStatement *reset_stmt = make_reset_array_base(s,dad,symb->type(),base, mHandle);
   s->insertStmtAfter(*reset_stmt, *s->controlParent());

     // pcrc_heap_alloc
   SgStatement *allocstmt=make_heap_alloc(s,symb->type(),base,dad,mHandle);
   s->insertStmtAfter(*allocstmt, *s->controlParent());

     // new DAD: pcrc_dad_a
   if (allocate ==1)
   {
	SgExpression *array = (SgExpression *)symb->attributeValue(0, Allocate_Def);
        ASSERT(array);
//	SgSymbol *symb = array->symbol();
	SgArrayType *type = isSgArrayType(symb->type());
        ASSERT(type);
	SgArrayType *arraytype = new SgArrayType(*type->baseType(),type->dimension(),*array->lhs());
	  // assume an array type as allocated.
	symb->setType(*arraytype);
	  // set_array_align  to each dimension
   	define_align_array(s, s, NULL);	// lastexec is NULL, means no need to insert pcrc_delete_array()
	  // recover the old type
	symb->setType(*type);
   }
  else
   	define_align_array(s, s, NULL);	// lastexec is NULL, means no need to insert pcrc_delete_array()
   
     // save pcrc_dad_a => pcrc_sdd_a
   if (allocate == 0)
   {
   SgExpression *sdadExp = new SgVarRefExp(sdad);
   SgExpression *dadExp = new SgVarRefExp(dad);
   SgStatement *saveStmt = new SgAssignStmt(*sdadExp, *dadExp);
   s->insertStmtAfter(*saveStmt, *s->controlParent());
   }

   s->deleteStmt();
}

void redistribute_array(SgStatement *s)
{
     // the symbol to distribute
   SgSymbol *symb = s->expr(0)->lhs()->symbol();
   int storage = getArrayStorage(symb);

	// redistribute a template
   if (symb->numberOfAttributes(Temp_Def) > 0) {
        Template_def * tdef = (Template_def *)symb->attributeValue(0, Temp_Def);
	ASSERT(tdef);
        SgSymbol * templatename= tdef->getsymbol();    
        Distribute_def * ddef = (Distribute_def *)s->attributeValue(0, Dist_Def);
	ASSERT(ddef);
        SgSymbol * rngname = s->getSymbolInCurrentScope(mixstring, "rng", templatename->identifier());
	distribute_template(s, rngname, tdef, ddef, s);
   }
   else 
	// redistribute an array
   {
   int allocate = 0;
   if (symb->numberOfAttributes(Allocate_Def)>0) allocate=1;

   SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", symb->identifier());
   SgSymbol * sdad = s->getSymbolInCurrentScope(mixstring, "sdd", symb->identifier());

   // create statements inversely

     // delete old DAD
   if (allocate == 0)
   {
   SgStatement *deleteStmt = make_delete_array(s, sdad);
   s->insertStmtAfter(*deleteStmt, *s->controlParent());

     // delete the allocated memory of old dad
   SgStatement *freeStmt = make_array_free(s, sdad);
   s->insertStmtAfter(*freeStmt, *s->controlParent());

     // add remap
   SgStatement *remap = make_remap(s, dad, sdad);
   s->insertStmtAfter(*remap, *s->controlParent());
   }

   SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "tbs", symb->identifier(), 0);
   SgSymbol * mHandle = s->getSymbolInCurrentScope(mixstring, "mh", symb->identifier(), 0);

    // reset_array_base
   SgStatement *reset_stmt = make_reset_array_base(s,dad,symb->type(),base, mHandle);
   s->insertStmtAfter(*reset_stmt, *s->controlParent());

     // pcrc_heap_alloc
   SgStatement *allocstmt = make_heap_alloc(s,symb->type(),base,dad,mHandle);
   s->insertStmtAfter(*allocstmt, *s->controlParent());

     // define dimensions
   Distribute_def * ddef = (Distribute_def *)s->attributeValue(0,Dist_Def);
   SgSymbol * procname = ddef->gettosymbol();
   Processor_def * pdef = (Processor_def *)procname->attributeValue(0, Proc_Def);
   SgExpression *array = NULL;
   if ( allocate == 1)
	array = (SgExpression *)symb->attributeValue(0, Allocate_Def);
   distribute_array(s, dad, array, ddef, pdef, s, AFTER);

   SgSymbol * procgrd = pdef->getgrp();
   SgExpression *procExp = new SgVarRefExp(procgrd);
   SgStatement *datastmt=make_new_array_data(s,dad,symb->type(),storage,-1,procExp);
   s->insertStmtAfter(*datastmt, *s->controlParent());

     // save pcrc_dad_a => pcrc_sdd_a
   if (allocate == 0)
   {
   SgExpression *sdadExp = new SgVarRefExp(sdad);
   SgExpression *dadExp = new SgVarRefExp(dad);
   SgStatement *saveStmt = new SgAssignStmt(*sdadExp, *dadExp);
   s->insertStmtAfter(*saveStmt, *s->controlParent());
   }

   }

   s->deleteStmt();
}

//----------------;

// add a common block
void add_global_common(SgStatement *s, SgStatement * lastdec)
{
  SgStatement *where = lastdec?lastdec:s;
  SgStatement *cp = lastdec?lastdec->controlParent():s;

  // add global common block
  SgExpression *expr_list = NULL;

  SgSymbol *globalcommon = s->getSymbolInCurrentScope("global");
  SgSymbol *incoresize = s->getSymbolInCurrentScope("IN_CORE_SIZE");
  SgExpression *incoresizeExp = new SgVarRefExp(incoresize);
  expr_list = new SgExpression(EXPR_LIST, incoresizeExp, expr_list, NULL);

  SgSymbol *recordsize = s->getSymbolInCurrentScope("RECORD_SIZE");
  SgExpression *recordsizeExp = new SgVarRefExp(recordsize);
  expr_list = new SgExpression(EXPR_LIST, recordsizeExp, expr_list, NULL);
  SgExpression *comm_list = new SgExpression(COMM_LIST, expr_list, NULL, globalcommon);

  SgStatement *comm_stmt = new SgStatement(COMM_STAT, NULL, NULL, comm_list, NULL, NULL);
  where->insertStmtAfter(*comm_stmt, *cp);
}

void init_global_common(SgStatement *s, SgStatement* lastdec)
{
  SgStatement *where = lastdec?lastdec:s;
  SgStatement *cp = lastdec?lastdec->controlParent():s;

  SgExpression *val;
  SgStatement *ass_stmt;

  {
  SgSymbol *recordsize = s->getSymbolInCurrentScope("RECORD_SIZE");
  SgExpression *recordsizeExp = new SgVarRefExp(recordsize);

  val = new SgValueExp(RECORD_SIZE);
  ass_stmt = new SgAssignStmt(*recordsizeExp, *val);
  
  where->insertStmtAfter(*ass_stmt, *cp);
  }

  {
  SgSymbol *incoresize = s->getSymbolInCurrentScope("IN_CORE_SIZE");
  SgExpression *incoresizeExp = new SgVarRefExp(incoresize);

  val = new SgValueExp(IN_CORE_SIZE);
  ass_stmt = new SgAssignStmt(*incoresizeExp, *val);
  
  where->insertStmtAfter(*ass_stmt, *cp);
  }

}

//----------------;

void add_definition(SgStatement * s)
{
  SgStatement * lastdec=s->lastDeclaration();

  SgStatement * lastexec; 
  // here we want the control-end or contains statement;
  for (int num=0; num < s->numberOfChildrenList1(); num++)
  {
      lastexec=s->childList1(num);
      if (lastexec->variant() == CONTAINS_STMT) break;
  }

  // define exit label if needed.
  // define_return_label(s);

  // add things in reverse order;
  allocate_unbound_array(s, lastdec, lastexec);

  lastdec=s->lastDeclaration();
  define_noalign_data(s, lastdec, lastexec);

  lastdec=s->lastDeclaration();
  define_align_array(s, lastdec, lastexec);
//  define_distribute_array(s, lastdec, lastexec);	// distribute without align

  // add some global variables
  lastdec=s->lastDeclaration();
  add_global_common(s, lastdec);

  lastdec=s->lastDeclaration();

  add_template(s, lastdec, lastexec);
  add_processor(s, lastdec, lastexec);

  if (in_main_program==on && s->variant() == PROG_HEDR)
  {
    add_init_finalize(s, lastdec, lastexec);
    init_global_common(s, lastdec);
  }

  if (in_procedure==on)
    add_stack_operation(s, lastdec, lastexec);
}

//----------------;

void add_var_startwith(SgProgHedrStmt * proghedr, char * string, SgType *type)
{
  SgSymbol * symb = proghedr->firstSymbolInScope();
  ASSERT(symb);
  SgExprListExp * declist = new SgExprListExp();
  
  for ( ; symb; symb=symb->nextInScope()) {
    if (!strncasecmp(symb->identifier(), string, strlen(string))) {
      // make sure this is a whole word
      if (strlen(symb->identifier()) > strlen(string) &&
	   symb->identifier()[strlen(string)] <= 'z' &&
	   symb->identifier()[strlen(string)] >= 'a' )
	continue;
      SgExpression * item = new SgVarRefExp(symb);
      declist->setLhs(*item);
      break;
    }
  }
  
  if (symb==NULL)
    return;
  else
    symb=symb->nextInScope();
  
  for ( ; symb; symb=symb->nextInScope()) {
    if (strncasecmp(symb->identifier(), string, strlen(string))==0) {
      if (strlen(symb->identifier()) > strlen(string) &&
	   symb->identifier()[strlen(string)] <= 'z' &&
	   symb->identifier()[strlen(string)] >= 'a' )
	continue;
      SgExpression * item = new SgVarRefExp (symb);
      SgExpression * temp = declist;
      while (temp->operand(2)!=NULL)
	temp=temp->operand(2);
      SgExpression * link = new SgExpression (EXPR_LIST, item, NULL, NULL);
      temp->setRhs(*link);
    }
  }
  
  if (declist->length()!=0) {
      SgVarDeclStmt * decstmt = new SgVarDeclStmt(*declist, *type);
      proghedr->childList1(0)->insertStmtBefore(*decstmt, *proghedr);
  }
}

//
void add_array_startwith(SgProgHedrStmt * proghedr, char * string, int size, SgType *type)
{
  SgSymbol * symb = proghedr->firstSymbolInScope();
  ASSERT(symb);
  SgExprListExp * declist = new SgExprListExp();
  
  for ( ; symb; symb=symb->nextInScope()) {
    if (!strncasecmp(symb->identifier(), string, strlen(string))) {
      if (strlen(symb->identifier()) > strlen(string) &&
	   symb->identifier()[strlen(string)] <= 'z' &&
	   symb->identifier()[strlen(string)] >= 'a' )
	continue;
      SgValueExp * dimExp=new SgValueExp(size);      
      SgTripletExp *ddot = new SgTripletExp(NULL,dimExp,NULL);
      SgArrayRefExp *item = new SgArrayRefExp(*symb);
      item->addSubscript(*ddot);
  
      declist->setLhs(*item);
      break;
    }
  }
  
  if (symb==NULL)
    return;
  else
    symb=symb->nextInScope();
  
  for ( ; symb; symb=symb->nextInScope()) {
    if (strncasecmp(symb->identifier(), string, strlen(string))==0) {
      if (strlen(symb->identifier()) > strlen(string) &&
	   symb->identifier()[strlen(string)] <= 'z' &&
	   symb->identifier()[strlen(string)] >= 'a' )
	continue;
      
      SgValueExp * dimExp=new SgValueExp(size);      
      SgTripletExp *ddot = new SgTripletExp(NULL,dimExp,NULL);
      SgArrayRefExp *item = new SgArrayRefExp(*symb);
      item->addSubscript(*ddot);

      SgExpression * temp = declist;
      while (temp->operand(2)!=NULL)
	temp=temp->operand(2);
      SgExpression * link = new SgExpression (EXPR_LIST, item, NULL, NULL);
      temp->setRhs(*link);
    }
  }
  
  if (declist->length()!=0) {
      SgVarDeclStmt * decstmt = new SgVarDeclStmt(*declist, *type);
	// zheng.gb
      SgStatement *stmt = proghedr->firstStmtForDeclaration();
      stmt->insertStmtAfter(*decstmt, *proghedr);
  }
}

//
void add_new_dec(SgStatement * s)
{
  SgProgHedrStmt * proghedr = (SgProgHedrStmt *) s;

  add_include(s);

  SgType *inttype = SgTypeInt();
  SgType *booltype = SgTypeBool();

    //define  pcrc_oc_stride_array(7),pcrc_oc_number(7) for
    // out of core read and write arguments
  add_array_startwith(proghedr, "pcrc_oc_number", 7, inttype);
  add_array_startwith(proghedr, "pcrc_oc_stride", 7, inttype);
  add_array_startwith(proghedr, "pcrc_oc_index", 7, inttype);  //slx
  add_array_startwith(proghedr, "pcrc_index", 7, inttype);  //slx
  add_array_startwith(proghedr, "pcrc_number", 7, inttype); //slx

  add_array_startwith(proghedr, "pcrc_gtl", 7, inttype);
  add_array_startwith(proghedr, "pcrc_gtu", 7, inttype);
  add_array_startwith(proghedr, "pcrc_a0",7, inttype);	// zheng.gb
  add_array_startwith(proghedr, "pcrc_a1",7, inttype);	// zheng.gb
  add_array_startwith(proghedr, "pcrc_b0",7, inttype);	// zheng.gb
  add_array_startwith(proghedr, "pcrc_b1",7, inttype);	// zheng.gb
  add_array_startwith(proghedr, "pcrc_c0",7, inttype);	// zheng.gb
  add_array_startwith(proghedr, "pcrc_c1",7, inttype);	// zheng.gb
  add_array_startwith(proghedr, "pcrc_ind0",7, inttype);	// zheng.gb
  add_array_startwith(proghedr, "pcrc_ind1",7, inttype);	// zheng.gb
  add_array_startwith(proghedr, "pcrc_amount",7, inttype);	// zheng.gb
  add_array_startwith(proghedr, "pcrc_tmp_rng",7, inttype);	// zheng.gb
   
  add_array_startwith(proghedr, "pcrc_arg_stride",7, inttype);	// huang
  add_var_startwith(proghedr, "pcrc_indepdo_rng", inttype);     // huang
  add_var_startwith(proghedr, "pcrc_dim_stride", inttype);     // huang
 
  // add out of core index variables
  add_var_startwith(proghedr, "pcrc_oci", inttype);
  add_var_startwith(proghedr, "pcrc_tmp_oci", inttype);
  add_var_startwith(proghedr, "pcrc_ST", inttype);
  
  add_var_startwith(proghedr, "pcrc_base", inttype);
  add_var_startwith(proghedr, "pcrc_comshift", booltype);	// zheng.gb
  add_var_startwith(proghedr, "pcrc_comm", inttype);	// zheng.gb
  add_var_startwith(proghedr, "pcrc_subs", inttype);	// zheng.gb for pcrc_gather
 
  add_var_startwith(proghedr, "pcrc_step", inttype);	// zheng.gb
  add_var_startwith(proghedr, "pcrc_mh", inttype);	// zheng.gb for MA_alloc

  add_var_startwith(proghedr, "pcrc_ast", inttype);
  add_var_startwith(proghedr, "pcrc_aub", inttype);
  add_var_startwith(proghedr, "pcrc_alb", inttype);

  add_var_startwith(proghedr, "pcrc_str", inttype);
  add_var_startwith(proghedr, "pcrc_sdx", inttype);
  add_var_startwith(proghedr, "pcrc_tmp_sdx", inttype);	// zheng.gb
  add_var_startwith(proghedr, "pcrc_stk_sdx", inttype);	// zheng.gb
  add_var_startwith(proghedr, "pcrc_T", inttype);	// slx
  add_var_startwith(proghedr, "pcrc_v", inttype);
  add_var_startwith(proghedr, "pcrc_u", inttype);

  add_var_startwith(proghedr, "pcrc_lil", inttype);
  add_var_startwith(proghedr, "pcrc_liu", inttype);
  add_var_startwith(proghedr, "pcrc_lis", inttype);
  add_var_startwith(proghedr, "pcrc_irg", inttype);

  add_var_startwith(proghedr, "pcrc_subgrp", inttype);	// zheng.gb
  add_var_startwith(proghedr, "pcrc_tdd", inttype);
  add_var_startwith(proghedr, "pcrc_tbs", inttype);
  add_var_startwith(proghedr, "pcrc_sdd", inttype);

  add_var_startwith(proghedr, "pcrc_dad", inttype);

  add_var_startwith(proghedr, "pcrc_glb", inttype); // lan

  add_var_startwith(proghedr, "pcrc_I", inttype);  // dbc
}



